class marketplaces:
    @staticmethod
    def US():
        url = "https://sellingpartnerapi-na.amazon.com"
        marketplace_id = "ATVPDKIKX0DER"
        return url, marketplace_id

    @staticmethod
    def CA():
        url = "https://sellingpartnerapi-na.amazon.com"
        marketplace_id = "A2EUQ1WTGCTBG2"
        return url, marketplace_id

    @staticmethod
    def MX():
        url = "https://sellingpartnerapi-na.amazon.com"
        marketplace_id = "A1AM78C64UM0Y8"
        return url, marketplace_id

    @staticmethod
    def BR():
        url = "https://sellingpartnerapi-na.amazon.com"
        marketplace_id = "A2Q3Y263D00KWC"
        return url, marketplace_id

    @staticmethod
    def UK():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A1F83G8C2ARO7P"
        return url, marketplace_id

    @staticmethod
    def DE():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A1PA6795UKMFR9"
        return url, marketplace_id

    @staticmethod
    def FR():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A13V1IB3VIYZZH"
        return url, marketplace_id

    @staticmethod
    def ES():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A1RKKUPIHCS9HS"
        return url, marketplace_id

    @staticmethod
    def IT():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "APJ6JRA9NG5V4"
        return url, marketplace_id

    @staticmethod
    def IN():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A21TJRUUN4KGV"
        return url, marketplace_id

    @staticmethod
    def JP():
        url = "https://sellingpartnerapi-fe.amazon.com"
        marketplace_id = "A1VC38T7YXB528"
        return url, marketplace_id

    @staticmethod
    def AU():
        url = "https://sellingpartnerapi-fe.amazon.com"
        marketplace_id = "A39IBJ37TRP1C6"
        return url, marketplace_id

    @staticmethod
    def AE():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A2VIGQ35RCS4UG"
        return url, marketplace_id

    @staticmethod
    def SG():
        url = "https://sellingpartnerapi-fe.amazon.com"
        marketplace_id = "A19VAU5U5O7RUS"
        return url, marketplace_id

    @staticmethod
    def TR():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A33AVAJ2PDY3EV"
        return url, marketplace_id

    @staticmethod
    def SA():
        url = "https://sellingpartnerapi-eu.amazon.com"
        marketplace_id = "A17E79C6D8DWNP"
        return url, marketplace_id
